; -*- fundamental -*-
;:
; Firmware for ABC80 SecureDigital controller
;:
; This needs a CPU mostly because ABC wants 256-byte sectors, and
; SecureDigital has 512-byte sectors, at least if one wants to have
; SDHC compatibility.
;:
; We mimic a "real" microcontroller-based design in the sense that we
; have separate ROM and RAM, the RAM being uninitalized.
; Total memory: 1K ROM, 2K RAM
; RAM:
; 8000-81FF -> controller data and stack
; 8200-83FF -> space for one SD sector
; 8400-87FF -> space for four ABC sectors (RAMBUFs)
;

	; -----------------------------------------------------------
	;  Data area
	; -----------------------------------------------------------
sysinfo:	equ 0x8000		; IY points here
command:	equ 0x8000		; Four byte controller command
sd_sect:	equ 0x8004		; LBA of SD sector in SD buffer
sd_need:	equ 0x8008		; SD sect needed (temp for load_buffer)
abc_buf:	equ 0x800C		; Address of selected RAMBUF
init_done:	equ 0x800E		; Initialization done flag
		; 0x800F unused

		; SD card argument buffers
zero_arg:	equ 0x8010		; All zero argument
sect_arg:	equ 0x8014		; Sector number as argument
card_ocr:	equ 0x8018		; Bit 30 is the SDHC flag
init_arg:	equ 0x801c		; Normally zero but modified by sd_init
clear_size:	equ 0x20
stack:		equ 0x8200		; Base of stack

ram_size:	equ 2048

sd_buf:		equ 0x8200
rambuf:		equ 0x8400

reset:					; Controller reset
	di
	im 1				; Interrupt = RST 38h
	ld sp,stack			; Reset stack
	jr reset_1

rst8:	jp sd_cmd			; rst8 = call sd_cmd
	nop
	nop
	nop
	nop
	nop

rst16:	xor a				; rst16 = DEHL <- 0, A <- 0
	ld l,a
	ld h,a
	ld e,a
	ld d,a
	ret

reset_1:
	ld hl,sysinfo
	push hl
	pop iy
	ld de,sysinfo+1
	ld bc,clear_size-1		; Size of data area
	ld (hl),l			; Zero
	ldir				; Clear the data area

do_interrupt:
	ld a,128
	out (0x00),a			; Signal NOT READY as quickly
	xor a
	out (0x01),a			; No auxilliary status
	ld sp,stack			; Reset stack
	ld iy,sysinfo			; Restore IY
	; Resume here to begin command
	jr idle

	defs 0x0038-$
interrupt_entry:			; Address 0038h
	jr do_interrupt

idle:
	ld e,0
idle_err:				; Errcode in E
	ei				; Allow interrupts
	; Prepare to receive command
setup_command:
	xor a
	out (0x00),a
	out (0x02),a
	push iy				; Address for input
	pop hl
	ld a,l
	out (0x04),a
	ld a,h
	out (0x05),a
	ld a,4
	out (0x02),a
	ld a,e
	or 0x81				; Ready for command
	out (0x00),a
wait_for_cmd:
	in a,(0x00)			; DMA status
	rrca
	jr c,wait_for_cmd
	jr have_command

	defs 0x0066-$
nmi_entry:				; Address 0066h
	di
	jp nmi

have_command:
	; We have received an entire command sequence

	;
	; Reset auxilliary status
	;
	xor a
	out (0x01),a

	; Do we need to initialize the card?  We do this here, rather
	; than at reset time, because ABC80 is pretty generous with bus
	; and controller resets, and wants response almost immediately
	; after.
	bit 0,(iy+0x0e)
	call z,sd_init

	;
	; Compute RAMBUF address
	;
	ld a,(iy+1)
	ld d,rambuf/(4*256)
	add a				; 2 high bits of A -> 2 low bits of D
	rl d
	add a
	rl d
	ld (iy+13),d			; Save pointer to addressed RAMBUF

do_command:
	;
	; READ SECTOR command
	;
	bit 0,(iy+0)
	jr z,not_read_sector

	call load_buffer		; Load the SD buffer with the addressed SD sector
	jr c,data_error

	call confirm			; Sector loaded successfully

	; Copy relevant half of SD buffer to controller buffer
	call bufsetup
	ldir				; Copy into controller buffer

not_read_sector:

	;
	; SECTOR TO HOST command
	;
	bit 1,(iy+0)			; SECTOR TO HOST
	jr z,not_sector_host
	ld c,0x07			; Controller->CPU
	call transfer
not_sector_host:

	;
	; SECTOR TO CONTROLLER command
	;
	bit 2,(iy+0)			; SECTOR TO CONTROLLER
	jr z,not_sector_controller
	ld c,0x05			; CPU->controller
	call transfer
not_sector_controller:

	;:
	; WRITE SECTOR command
	;:
	bit 3,(iy+0)
	jr z,not_write_sector

	; First, load the appropriate SD sector
	call load_buffer
	jr c,data_error

	; Copy controller buffer to relevant half of SD buffer
	call bufsetup
	ex de,hl			; Controller buffer -> SD buffer
	ldir				; Copy into controller buffer

	; Finally, write back buffer to SD card
	call write_buffer
	jr c,data_error

	call confirm			; sd_sect up to date with disk

not_write_sector:

	;
	; All other commands ignored
	;
	jp idle

	;
	; Place to jump when an error is detected; auxilliary error
	; code should be placed in register E
	;
data_error:
	ld a,e				; Auxilliary status code
	out (0x01),a
	ld e,0x08			; ERROR
	jp idle_err

	;:
	; Set up register for a SD buffer -> controller buffer
	; transfer for the currently addressed buffer and sector.
	; For the opposite direction transfer, do ex de,hl before
	; actually transferring data
	;
bufsetup:
	ld hl,sd_buf/2
	ld bc,256			; 256 bytes
	ld a,(iy+3)			; Sector 
	rrca
	rl h				; HL -> proper half of SD buffer
	ld de,(abc_buf)
	ret
	
	;:
	; Transfer a sector to or from the controller.  Register C
	; should be set to 0x05 for host->controller and 0x07 for
	; controller->host
	;
transfer:
	xor a				; Controller buffer always aligned
	out (0x04),a
	ld a,(iy+13)			; High byte of addressed controller buffer
	out (c),a
	xor a
	out (0x03),a			; 256 bytes
	inc a				; A = 01h (READY)
	out (0x00),a
tr_wait:
	in a,(0x00)			; DMA engine status
	rrca				; Bit 0 = DMA done
	ret nc
	jr tr_wait
	
;:
; This routine converts the command contents to a SD sector number,
; and loads it if it is not currently loaded.
;
load_buffer:
	; Initialize buffer addresses
	xor a
	ld (iy+10),a
	ld (iy+11),a

	; Get track
	ld e,a				; Seek error
	ld a,(iy+2)
	cp 238				; 238 tracks/volume
	ccf
	ret c				; Bad sector 
	srl a
	ld (iy+9),a

	; Get sector
	ld a,(iy+3)
	rr a				; Convert to SD sector + low byte of track
	ld (iy+8),a

	; Compute unit number -> third LBA byte
	ld a,(iy+1)			; K1 = unit  + more
	and 0x07			; Unit 
	jr z,lb_haveunit		; Unit 0 skips loop
	; Each unit has 238 tracks, that is the maximum supported.
	; We multiply by 119 (238/2), this starts LBA bytes 2-3
	ld b,a
lb_getunit:
	ld a,(iy+9)
	add a,119
	ld (iy+9),a
	ld a,(iy+10)
	adc a,0
	ld (iy+10),a
	djnz lb_getunit
lb_haveunit:

	; Now sd_need contains the LBA sector (including upper 4 control bits)
	; that we need; sd_sect contains what's currently loaded in the SD
	; sector buffer.
	ld a,(iy+8)
	cp (iy+4)
	jr nz,lb_nomatch
	ld a,(iy+9)
	cp (iy+5)
	jr nz,lb_nomatch
	ld a,(iy+10)
	cp (iy+6)
	jr nz,lb_nomatch
	ld a,(iy+11)
	cp (iy+7)
	ret z				; If match, we're done

lb_nomatch:
	; No such luck.  Need to actually load the sector.
	call invalidate			; We're about to overwrite sd_buf
	call fmt_lba			; Convert to byte count if appropriate
	ld a,17				; CMD17 - READ_SINGLE_BLOCK
	di				; Don't confuse the card...
	rst 8
	and a
	jr nz,bad_card

lb_wait_data:
	in a,(0x34)
	inc a
	jr z,lb_wait_data		; -1 = no response
	inc a
	jr nz,lb_error			; -2 = data token; other is error

	ld hl,sd_buf
	ld bc,0ff30h			; 255 bytes from port 30h
	inir
	inir				; ... plus 256 bytes

	; At this point, the last byte is latched, and the CRC16 is valid

	in a,(0x22)			; CRC16 MSB
	ld d,a
	in a,(0x23)			; CRC16 LSB
	ld e,a

	ini				; Store the last data byte

	in a,(0x30)			; CRC16 MSB
	xor d
	ld d,a
	in a,(0x30)			; CRC16 LSB
	ei
	xor e
	or d
	jr nz,lb_error

	ret				; F.C = 0 from "or d"

lb_error:
	; Read CRC error
	scf
	ld e,0x08			; Read error
	ret

bad_card:
	jp nmi

;:
; Write the buffer currently in the SD sector buffer back to the SD card
;
write_buffer:
	in a,(0x00)
	and 08h				; Write Protect
	ld e,040h			; Write Protect error
	jr nz,wb_error
	call sd_status			; Clear the card status
	call fmt_lba			; Set up the LBA
	ld a,24				; CMD24 - WRITE_BLOCK
	di				; Don't confuse the card...
	rst 8
	and a
	jr nz,bad_card

	ld a,0FEh			; Data token
	out (0x3d),a
	ld bc,0x3d			; 256 bytes from port 0x3d
	ld hl,sd_buf
	outi				; Send first byte and clear CRC regs
	ld c,0x39
	otir
	otir

	in a,(0x23)			; CRC16 LSB
	ld c,a
	in a,(0x22)			; CRC16 MSB
	out (0x39),a
	ld a,c
	out (0x39),a

	in a,(0x30)			; Clock out the response token
	in a,(0x20)
	ld c,a				; Save response flag

wb_wait_unbusy:
	in a,(0x30)
	inc a
	jr nz,wb_wait_unbusy
	ei				; Transaction complete (enough)

	ld a,c
	and 1Fh
	cp 05h
	jr nz,wb_error_20

	; Card got at least to the programming phase.  We need to get
	; the status register from CMD13 to know if the programming
	; went okay.
	call sd_status
	ld d,a
	and a
	jr nz,wb_error_20
	ld e,a
	and a
	ret z				; No error, we're done
	ld e,10h			; Seek error
	jp m,wb_error
	ld d,a
	and 23h
	ld e,40h			; Write Protect error
	jr nz,wb_error
	ld e,08h			; CRC error
	bit 4,d
	jr nz,wb_error
wb_error_20:
	ld e,20h			; Head load/write fault
wb_error:
	call invalidate			; We don't have a valid buffer...
	scf
	ret

;
; Format an LBA (in sd_need) appropriately, setting DEHL appropriately
; for a read or write command.
;
; If we are in SDHC mode, just load it; if we are not, convert to a
; byte offset.
;
fmt_lba:
	bit 6,(iy+0x17)
	jr nz,fmt_lba_sdhc

	; Format the LBA for non-SDHC mode
	ld l,0
	ld h,(iy+8)
	ld de,(sd_need+1)

	sla h
	rl e
	rl d
	ret

fmt_lba_sdhc:
	ld hl,(sd_need)
	ld de,(sd_need+2)
	ret

;:
; NMI handing -- we receive an NMI when the SD card is removed.  Allow
; a DMA operation in process to complete, then report CMD/ERROR/READY
; with the error code of DRIVE NOT READY (80h) until the card is reinserted.
;:
; We do a full initialization after a card has been inserted; this includes
; invalidating the SD sector buffer!
;
nmi:
nmi_wait_dma:
	di
	xor a
	out (0x82),a			; Disable DMA

	ld a,0x08			; Drive not ready
	out (0x81),a			; Auxilliary status
	xor a
	out (0x80),a			; Primary status

nmi_wait_card:
	in a,(0x80)			; DMA/SD aux status
	and 04h				; Card present?
	jr z,nmi_wait_card

	rst 0				; Reset everything

;
; Initialize the SD card after power on or insert
;
sd_init:
	call invalidate			; Invalidate sd_sect
	rst 16				; DEHL <- 0
	;
	; Allow card to start up
	;
	ld a,3				; Low speed, CS# high
	out (0x22),a
	ld b,0				; 256 loops
	ld a,0xFF
sd_init_wait:
	out (0x31),a			; Send FF
	djnz sd_init_wait
	;
	; Reset card into SPI mode
	;
	ld a,1
	out (0x22),a			; Low speed, CS# low
	; command is already zero
	xor a
	rst 8				; CMD0 - GO_IDLE_STATE
	dec a
	jr nz,sd_init			; Not right?  Try again
	;
	; Enable command CRC checking, if available.
	;
	out (0x22),a			; A = 0: High speed, CS# low
	inc l				; Enable CRC flag
	ld a,59				; CMD59 - CDC_ON_OFF
	rst 8
	; If this command fails, ignore it.
	dec l				; Restore DEHL <- 1
	;
	; Probe for extended features
	;
	ld hl,0x01aa			; Check pattern, 3.3 V support
	ld a,8				; CMD8 - SEND_IF_COND
	rst 8
	bit 2,a
	jr nz,old_card			; CMD8 not supported
	dec a
	jr nz,sd_carderr		; Other error
	in a,(0x30)			; Command version
	in a,(0x30)			; Reserved bits
	in a,(0x30)			; Voltage range accepted
	sub h
	ld h,a
	in a,(0x30)			; Check pattern
	sub l
	or h
	jr nz,sd_carderr
	ld l,a				; Restore DEHL = 0
	;
	; Initialize card
	;
	ld d,40h			; SDHC supported
old_card:
sd_init_acmd41:
	ld a,41
	call sd_acmd
	bit 2,a
	jr nz,sd_init_cmd1
	and a
	jr z,sd_probe_ocr
	dec a
	jr z,sd_init_acmd41
	jr sd_carderr

sd_init_cmd1:
	xor a
	ld d,a				; DEHL <- 0
	inc a
	rst 8				; CMD1 - SEND_OP_COND
	and a
	jr z,sd_set_blklen
	dec a
	jr z,sd_init_cmd1
	jr sd_carderr

	;
	; Probe the OCR for SHDC mode
	;
sd_probe_ocr:
	; A = 0 here
	ld (iy+0x17),a			; Assume no SDHC
	bit 6,d				; Did we not even offer SDHC?
	jr z,no_sdhc
	ld d,a				; DEHL <- 0
	ld a,58				; CMD58 - GET_OCR
	rst 8
	and a
	jr nz,sd_carderr
	push hl
	ld hl,card_ocr+3
	ld bc,0x0430			; 4 bytes from port 30h
	otdr
	pop hl
no_sdhc:

	;
	; Set block length to 512 bytes (2-4 GB cards sometimes
	; boot up defaulting to a longer block size, even though
	; that violates the spec.)
	;
sd_set_blklen:
	ld h,2				; Argument <- 512
	ld a,16				; CMD16 - SET_BLOCKLEN
	rst 8
	and a
	jr nz,sd_carderr

sd_done:
	set 0,(iy+0x0e)			; Mark initialization done
	ret

;
; We end up here if we got an error during initialization.
; Room for better error handling...
;
sd_carderr:
	jp sd_init

;
; Issue a SD card application command and obtain the response.
; The command number is in register A, and the 32-bit argument
; is in registers DEHL.
; Returns the (first byte of) the response code in A, all other
; registers are preserved.
;
sd_acmd:
	ex af,af'			; Save command number
	exx				; Save argument registers DEHL
	rst 16				; DEHL <- 0
	ld a,55				; CMD55 - APP_CMD
	rst 8
	exx
	bit 2,a				; Unknown command (very old card)?
	jr nz,no_acmd
	and 0FEh
	jp nz,nmi			; APP_CMD shouldn't error...
	ex af,af'
	; Fall through

;
; Issue an SD card command and obtain the response.
; The command number is in register A, and the 32-bit argument
; is in registers DEHL.
; Returns the (first byte of) the response code in A, all other
; registers are preserved.
;
; Can be invoked with RST 8.
;
sd_cmd:
	push bc
	or 0x40				; Make a command token
	ld bc,0x1039			; Port 0x39; max 16 cycles before error
	out (0x3d),a			; Send command code, clear CRC
	out (c),d			; Send argument in bigendian order
	out (c),e
	out (c),h
	out (c),l
	in a,(0x29)			; Get CRC7
	out (0x39),a			; Send CRC7
sd_cmd_resp:
	in a,(0x30)
	and a
	jp p,sd_cmd_out
	djnz sd_cmd_resp
	jp nmi				; Card not responding, assume removed
sd_cmd_out:
	pop bc
no_acmd:
	ret

;
; Execute the status command (CMD13), returning the 16-bit status in DE
; A is clobbered.
;
sd_status:
	push hl
	di
	rst 16				; DEHL <- 0
	ld a,13				; CMD13 - GET_STATUS
	rst 8
	ld d,a
	in a,(0x30)
	ei
	ld e,a
	pop hl
	ret

;
; Mark sd_sect invalid.
;
invalidate:
	ld a,0xff
	ld (iy+4),a
	ld (iy+5),a
	ld (iy+6),a
	ld (iy+7),a
	ret

;
; Mark sd_sect up to date with the disk sector in sd_need.
;
confirm:
	ld hl,sd_need
	ld de,sd_sect
	ld bc,4
	ldir
	ret
